package com.b510.big.data.storm.process;

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import storm.kafka.KafkaSpout;
import storm.kafka.SpoutConfig;
import storm.kafka.StringScheme;
import storm.kafka.ZkHosts;
import storm.kafka.bolt.KafkaBolt;
import storm.kafka.bolt.mapper.FieldNameBasedTupleToKafkaMapper;
import storm.kafka.bolt.selector.DefaultTopicSelector;
import backtype.storm.Config;
import backtype.storm.LocalCluster;
import backtype.storm.StormSubmitter;
import backtype.storm.generated.AlreadyAliveException;
import backtype.storm.generated.InvalidTopologyException;
import backtype.storm.spout.SchemeAsMultiScheme;
import backtype.storm.topology.BasicOutputCollector;
import backtype.storm.topology.OutputFieldsDeclarer;
import backtype.storm.topology.TopologyBuilder;
import backtype.storm.topology.base.BaseBasicBolt;
import backtype.storm.tuple.Fields;
import backtype.storm.tuple.Tuple;
import backtype.storm.tuple.Values;

public class LogFilterTopology {

	public static void main(String[] args) {

		ZkHosts zkHosts = new ZkHosts(Common.ZOOKEEPER_QUORUM);
		//Spout从'filtered_log' topic里面获取数据
		SpoutConfig spoutConfig = new SpoutConfig(zkHosts, Common.ALL_MY_LOG_TOPIC, Common.ZOOKEEPER_ROOT, Common.ZOOKEEPER_ID);
		List<String> zkServers = new ArrayList<>();
		for (String host : zkHosts.brokerZkStr.split(",")) {
			zkServers.add(host.split(":")[0]);
		}

		spoutConfig.zkServers = zkServers;
		spoutConfig.zkPort = Common.ZOOKEEPER_PORT;
		spoutConfig.forceFromStart = true;
		spoutConfig.socketTimeoutMs = 60 * 60 * 1000;
		spoutConfig.scheme = new SchemeAsMultiScheme(new StringScheme());

		// 创建KafkaSpout
		KafkaSpout kafkaSpout = new KafkaSpout(spoutConfig);

		TopologyBuilder builder = new TopologyBuilder();
		// Storm从Kafka消费数据
		builder.setSpout(Common.KAFKA_SPOUT, kafkaSpout, 3);
		// Storm对数据进行筛选(Buy-已经购买,Shopping_Car-潜在购买)
		builder.setBolt(Common.FILTER_BOLT, new FilterBolt(), 8).shuffleGrouping(Common.KAFKA_SPOUT);

		// 创建KafkaBolt
		@SuppressWarnings({ "unchecked", "rawtypes" })
		KafkaBolt kafkaBolt = new KafkaBolt().withTopicSelector(new DefaultTopicSelector(Common.FILTERED_LOG_TOPIC)).withTupleToKafkaMapper(new FieldNameBasedTupleToKafkaMapper());

		// Storm把筛选的数据放入到Kafka
		builder.setBolt(Common.KAFKA_BOLT, kafkaBolt, 2).shuffleGrouping(Common.FILTER_BOLT);

		Properties props = new Properties();
		props.put("metadata.broker.list", Common.STORM_METADATA_BROKER_LIST);
		props.put("request.required.acks", Common.STORM_REQUEST_REQUIRED_ACKS);
		props.put("serializer.class", Common.STORM_SERILIZER_CLASS);

		Config conf = new Config();
		conf.put("kafka.broker.properties", props);

		conf.put(Config.STORM_ZOOKEEPER_SERVERS, zkServers);

		if (args == null || args.length == 0) {
			// 本地方式运行
			LocalCluster localCluster = new LocalCluster();
			localCluster.submitTopology("storm-kafka-topology", conf, builder.createTopology());
		} else {
			// 集群方式运行
			conf.setNumWorkers(3);
			try {
				StormSubmitter.submitTopology(args[0], conf, builder.createTopology());
			} catch (AlreadyAliveException | InvalidTopologyException e) {
				System.out.println("error : " + e);
			}
		}
	}
}

class FilterBolt extends BaseBasicBolt {

	private static final long serialVersionUID = 1L;

	@Override
	public void execute(Tuple input, BasicOutputCollector collector) {
		String logStr = input.getString(0);
		// 只针对我们感兴趣的关键字进行过滤
		// 这里我们过滤包含'Buy', 'Shopping_Car'的日志信息
		if (logStr.contains(Common.KEY_WORD_BUY) || logStr.contains(Common.KEY_WORD_SHOPPING_CAR)) {
			collector.emit(new Values(logStr));
		}
	}

	@Override
	public void declareOutputFields(OutputFieldsDeclarer declarer) {
		declarer.declare(new Fields(FieldNameBasedTupleToKafkaMapper.BOLT_MESSAGE));
	}
}

class Common {
	public static final String ALL_MY_LOG_TOPIC = "all_my_log";
	public static final String FILTERED_LOG_TOPIC = "filtered_log";
	
	public static final String DATE_FORMAT_YYYYDDMMHHMMSS = "yyyyMMddHHmmss";
	public static final String DATE_FORMAT_HHMMSS = "HHmmss";
	public static final String DATE_FORMAT_HHMMSS_DEFAULT_VALUE = "000001";

	public static final String HBASE_ZOOKEEPER_LIST = "node1:2888,node2:2888,node3:2888";
	public static final int ZOOKEEPER_PORT = 2181;
	public static final String ZOOKEEPER_QUORUM = "node1:" + ZOOKEEPER_PORT + ",node2:" + ZOOKEEPER_PORT + ",node3:" + ZOOKEEPER_PORT + "";
	public static final String ZOOKEEPER_ROOT = "/MyKafka";
	public static final String ZOOKEEPER_ID = "MyTrack";

	public static final String KAFKA_SPOUT = "kafkaSpout";
	public static final String FILTER_BOLT = "filterBolt";
	public static final String PROCESS_BOLT = "processBolt";
	public static final String HBASE_BOLT = "hbaseBolt";
	public static final String KAFKA_BOLT = "kafkaBolt";

	// Storm Conf
	public static final String STORM_METADATA_BROKER_LIST = "node1:9092,node2:9092,node3:9092";
	public static final String STORM_REQUEST_REQUIRED_ACKS = "1";
	public static final String STORM_SERILIZER_CLASS = "kafka.serializer.StringEncoder";

	// key word
	public static final String KEY_WORD_BUY = "Buy";
	public static final String KEY_WORD_SHOPPING_CAR = "Shopping_Car";
	
	//hbase
	public static final String TABLE_USER_ACTION = "t_user_actions";
	public static final String COLUMN_FAMILY = "cf";
	//间隔多少秒写入Hbase一次
	public static final int WRITE_RECORD_TO_TABLE_PER_SECOND = 1;
	//需要记录5年的数据
	public static final int TABLE_MAX_VERSION = (60/WRITE_RECORD_TO_TABLE_PER_SECOND) * 60 * 24 * 365 * 5;
}

