package com.b510.big.data.storm.process;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hbase.HColumnDescriptor;
import org.apache.hadoop.hbase.HTableDescriptor;
import org.apache.hadoop.hbase.TableName;
import org.apache.hadoop.hbase.client.HBaseAdmin;
import org.apache.hadoop.hbase.client.HConnection;
import org.apache.hadoop.hbase.client.HConnectionManager;
import org.apache.hadoop.hbase.client.HTableInterface;
import org.apache.hadoop.hbase.client.Put;

import storm.kafka.KafkaSpout;
import storm.kafka.SpoutConfig;
import storm.kafka.StringScheme;
import storm.kafka.ZkHosts;
import backtype.storm.Config;
import backtype.storm.LocalCluster;
import backtype.storm.StormSubmitter;
import backtype.storm.generated.AlreadyAliveException;
import backtype.storm.generated.InvalidTopologyException;
import backtype.storm.spout.SchemeAsMultiScheme;
import backtype.storm.task.TopologyContext;
import backtype.storm.topology.BasicOutputCollector;
import backtype.storm.topology.IBasicBolt;
import backtype.storm.topology.OutputFieldsDeclarer;
import backtype.storm.topology.TopologyBuilder;
import backtype.storm.topology.base.BaseBasicBolt;
import backtype.storm.tuple.Fields;
import backtype.storm.tuple.Tuple;
import backtype.storm.tuple.Values;

public class LogProcessTopology {

	public static void main(String[] args) {

		ZkHosts zkHosts = new ZkHosts(Common.ZOOKEEPER_QUORUM);
		//Spout从'filtered_log' topic里面获取数据
		SpoutConfig spoutConfig = new SpoutConfig(zkHosts, Common.FILTERED_LOG_TOPIC, Common.ZOOKEEPER_ROOT, Common.ZOOKEEPER_ID);
		List<String> zkServers = new ArrayList<>();
		for (String host : zkHosts.brokerZkStr.split(",")) {
			zkServers.add(host.split(":")[0]);
		}

		spoutConfig.zkServers = zkServers;
		spoutConfig.zkPort = Common.ZOOKEEPER_PORT;
		spoutConfig.forceFromStart = true;
		spoutConfig.socketTimeoutMs = 60 * 60 * 1000;
		spoutConfig.scheme = new SchemeAsMultiScheme(new StringScheme());

		// 创建KafkaSpout
		KafkaSpout kafkaSpout = new KafkaSpout(spoutConfig);

		TopologyBuilder builder = new TopologyBuilder();
		// Storm再次从Kafka消费数据
		builder.setSpout(Common.KAFKA_SPOUT, kafkaSpout, 3);
		// Storm对数据进行统计（Buy-已经购买人数,Shopping_Car-潜在购买人数）
		builder.setBolt(Common.PROCESS_BOLT, new ProcessBolt(), 3).shuffleGrouping(Common.KAFKA_SPOUT);
		// Storm将数据写入到Hbase
		builder.setBolt(Common.HBASE_BOLT, new HbaseBolt(), 3).shuffleGrouping(Common.PROCESS_BOLT);

		Properties props = new Properties();
		props.put("metadata.broker.list", Common.STORM_METADATA_BROKER_LIST);
		props.put("request.required.acks", Common.STORM_REQUEST_REQUIRED_ACKS);
		props.put("serializer.class", Common.STORM_SERILIZER_CLASS);

		Config conf = new Config();
		conf.put("kafka.broker.properties", props);

		conf.put(Config.STORM_ZOOKEEPER_SERVERS, zkServers);

		if (args == null || args.length == 0) {
			// 本地方式运行
			LocalCluster localCluster = new LocalCluster();
			localCluster.submitTopology("storm-kafka-topology", conf, builder.createTopology());
		} else {
			// 集群方式运行
			conf.setNumWorkers(3);
			try {
				StormSubmitter.submitTopology(args[0], conf, builder.createTopology());
			} catch (AlreadyAliveException | InvalidTopologyException e) {
				System.out.println("error : " + e);
			}
		}
		
	}
}

class ProcessBolt extends BaseBasicBolt {

	private static final long serialVersionUID = 1L;

	@Override
	public void execute(Tuple input, BasicOutputCollector collector) {
		String logStr = input.getString(0);
		if (logStr != null) {
			String infos[] = logStr.split("\\t");
			//180.21.100.66	贵州	2018-12-20	1545290594752	5270357330431599426	www.hongten.com	Buy
			collector.emit(new Values(infos[2], infos[6]));
		}
	}

	@Override
	public void declareOutputFields(OutputFieldsDeclarer declarer) {
		declarer.declare(new Fields("date", "user_action"));
	}
}

class HbaseBolt implements IBasicBolt {
	private static final long serialVersionUID = 1L;

	HBaseDAO hBaseDAO = null;
	
	SimpleDateFormat simpleDateFormat = null;
	SimpleDateFormat simpleDateFormatHHMMSS = null;
	
	int userBuyCount = 0;
	int userShoopingCarCount = 0;
	
	//这里要考虑避免频繁写入数据到hbase
	int writeToHbaseMaxNum = Common.WRITE_RECORD_TO_TABLE_PER_SECOND * 1000;
	long begin = System.currentTimeMillis();
	long end = 0;
	
	@SuppressWarnings("rawtypes")
	@Override
	public void prepare(Map map, TopologyContext context) {
		hBaseDAO = new HBaseDAOImpl();
		simpleDateFormat = new SimpleDateFormat(Common.DATE_FORMAT_YYYYDDMMHHMMSS);
		simpleDateFormatHHMMSS = new SimpleDateFormat(Common.DATE_FORMAT_HHMMSS);
		hBaseDAO.createTable(Common.TABLE_USER_ACTION, new String[]{Common.COLUMN_FAMILY}, Common.TABLE_MAX_VERSION);
	}

	@Override
	public void execute(Tuple input, BasicOutputCollector collector) {
		// 如果时间是第二天的凌晨1s
		// 需要对count做清零处理
		//不过这里的判断不是很准确，因为在此时，可能前一天的数据还没有处理完
		if (simpleDateFormatHHMMSS.format(new Date()).equals(Common.DATE_FORMAT_HHMMSS_DEFAULT_VALUE)) {
			userBuyCount = 0;
			userShoopingCarCount = 0;
		}
		
		if (input != null) {
			// base one ProcessBolt.declareOutputFields()
			String date = input.getString(0);
			String userAction = input.getString(1);

			if (userAction.equals(Common.KEY_WORD_BUY)) {
				//同一个user在一天之内可以重复'Buy'动作
				userBuyCount++;
			}

			if (userAction.equals(Common.KEY_WORD_SHOPPING_CAR)) {
				userShoopingCarCount++;
			}

			end = System.currentTimeMillis();
			if ((end - begin) > writeToHbaseMaxNum) {
				System.out.println("hbase_key: " + Common.KEY_WORD_BUY + "_" + date + " , userBuyCount: " + userBuyCount + ", userShoopingCarCount :" + userShoopingCarCount);
				
				//往hbase中写入数据
				String quailifer = simpleDateFormat.format(new Date());
				hBaseDAO.insert(Common.TABLE_USER_ACTION , 
						Common.KEY_WORD_BUY + "_" + date, 
						Common.COLUMN_FAMILY, 
						new String[] { quailifer },
						new String[] { "{user_buy_count:" + userBuyCount + "}" }
						);
				hBaseDAO.insert(Common.TABLE_USER_ACTION , 
						Common.KEY_WORD_SHOPPING_CAR + "_" + date, 
						Common.COLUMN_FAMILY, 
						new String[] { quailifer },
						new String[] { "{user_shopping_car_count:" + userShoopingCarCount + "}" }
						);
				begin = System.currentTimeMillis();
			}
		}
	}

	@Override
	public void declareOutputFields(OutputFieldsDeclarer declarer) {

	}

	@Override
	public Map<String, Object> getComponentConfiguration() {
		return null;
	}

	@Override
	public void cleanup() {

	}
}

interface HBaseDAO {
	public void createTable(String tableName, String[] columnFamilys, int maxVersion);
	public void insert(String tableName, String rowKey, String family, String quailifer[], String value[]);
}

class HBaseDAOImpl implements HBaseDAO {

	HConnection hConnection = null;
	static Configuration conf = null;

	public HBaseDAOImpl() {
		conf = new Configuration();
		conf.set("hbase.zookeeper.quorum", Common.HBASE_ZOOKEEPER_LIST);
		try {
			hConnection = HConnectionManager.createConnection(conf);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void createTable(String tableName, String[] columnFamilys, int maxVersion) {
		try {
			HBaseAdmin admin = new HBaseAdmin(conf);
			if (admin.tableExists(tableName)) {
				System.err.println("table existing in hbase.");
			} else {
				HTableDescriptor tableDesc = new HTableDescriptor(TableName.valueOf(tableName));
				for (String columnFamily : columnFamilys) {
					HColumnDescriptor hColumnDescriptor = new HColumnDescriptor(columnFamily);
					hColumnDescriptor.setMaxVersions(maxVersion);
					tableDesc.addFamily(hColumnDescriptor);
				}

				admin.createTable(tableDesc);
				System.err.println("table is created.");
			}
			admin.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@Override
	public void insert(String tableName, String rowKey, String family, String quailifer[], String value[]) {
		HTableInterface table = null;
		try {
			table = hConnection.getTable(tableName);
			Put put = new Put(rowKey.getBytes());
			for (int i = 0; i < quailifer.length; i++) {
				String col = quailifer[i];
				String val = value[i];
				put.add(family.getBytes(), col.getBytes(), val.getBytes());
			}
			table.put(put);
			System.err.println("save record successfuly.");
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			try {
				table.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

}

